"""
Routes for interacting with log objects.
"""

from typing import Optional, Sequence

from fastapi import Body, Depends, WebSocket, status
from pydantic import TypeAdapter
from starlette.status import WS_1002_PROTOCOL_ERROR

import prefect.server.api.dependencies as dependencies
import prefect.server.models as models
from prefect.server.database import PrefectDBInterface, provide_database_interface
from prefect.server.logs import stream
from prefect.server.schemas.actions import LogCreate
from prefect.server.schemas.core import Log
from prefect.server.schemas.filters import LogFilter
from prefect.server.schemas.sorting import LogSort
from prefect.server.utilities import subscriptions
from prefect.server.utilities.server import PrefectRouter

router: PrefectRouter = PrefectRouter(prefix="/logs", tags=["Logs"])


@router.post("/", status_code=status.HTTP_201_CREATED)
async def create_logs(
    logs: Sequence[LogCreate],
    db: PrefectDBInterface = Depends(provide_database_interface),
) -> None:
    """
    Create new logs from the provided schema.

    For more information, see https://docs.prefect.io/v3/how-to-guides/workflows/add-logging.
    """
    for batch in models.logs.split_logs_into_batches(logs):
        async with db.session_context(begin_transaction=True) as session:
            await models.logs.create_logs(session=session, logs=batch)


logs_adapter: TypeAdapter[Sequence[Log]] = TypeAdapter(Sequence[Log])


@router.post("/filter")
async def read_logs(
    limit: int = dependencies.LimitBody(),
    offset: int = Body(0, ge=0),
    logs: Optional[LogFilter] = None,
    sort: LogSort = Body(LogSort.TIMESTAMP_ASC),
    db: PrefectDBInterface = Depends(provide_database_interface),
) -> Sequence[Log]:
    """
    Query for logs.
    """
    async with db.session_context() as session:
        return logs_adapter.validate_python(
            await models.logs.read_logs(
                session=session, log_filter=logs, offset=offset, limit=limit, sort=sort
            )
        )


@router.websocket("/out")
async def stream_logs_out(websocket: WebSocket) -> None:
    """Serve a WebSocket to stream live logs"""
    websocket = await subscriptions.accept_prefect_socket(websocket)
    if not websocket:
        return

    try:
        # After authentication, the next message is expected to be a filter message, any
        # other type of message will close the connection.
        message = await websocket.receive_json()

        if message["type"] != "filter":
            return await websocket.close(
                WS_1002_PROTOCOL_ERROR, reason="Expected 'filter' message"
            )

        try:
            filter = LogFilter.model_validate(message["filter"])
        except Exception as e:
            return await websocket.close(
                WS_1002_PROTOCOL_ERROR, reason=f"Invalid filter: {e}"
            )

        # No backfill support for logs - only live streaming
        # Subscribe to the ongoing log stream
        async with stream.logs(filter) as log_stream:
            async for log in log_stream:
                if not log:
                    if await subscriptions.still_connected(websocket):
                        continue
                    break

                await websocket.send_json(
                    {"type": "log", "log": log.model_dump(mode="json")}
                )

    except subscriptions.NORMAL_DISCONNECT_EXCEPTIONS:  # pragma: no cover
        pass  # it's fine if a client disconnects either normally or abnormally

    return None



"""
Version command — native cyclopts implementation.

Displays detailed version and integration information.
"""

import platform
import sqlite3
import sys
from importlib.metadata import PackageNotFoundError
from importlib.metadata import version as import_version
from typing import Annotated, Any

import cyclopts

import prefect
import prefect.cli._app as _cli
import prefect.context
from prefect.cli._utilities import with_cli_exception_handling


@with_cli_exception_handling
async def version(
    *,
    omit_integrations: Annotated[
        bool,
        cyclopts.Parameter("--omit-integrations", help="Omit integration information"),
    ] = False,
):
    """Get the current Prefect version and integration information."""
    from prefect.client.base import determine_server_type
    from prefect.client.constants import SERVER_API_VERSION
    from prefect.settings import get_current_settings
    from prefect.types._datetime import parse_datetime

    if build_date_str := prefect.__version_info__.get("date", None):
        build_date = parse_datetime(build_date_str).strftime("%a, %b %d, %Y %I:%M %p")
    else:
        build_date = "unknown"

    version_info: dict[str, Any] = {
        "Version": prefect.__version__,
        "API version": SERVER_API_VERSION,
        "Python version": platform.python_version(),
        "Git commit": prefect.__version_info__["full-revisionid"][:8],
        "Built": build_date,
        "OS/Arch": f"{sys.platform}/{platform.machine()}",
        "Profile": prefect.context.get_settings_context().profile.name,
    }
    server_type = determine_server_type()

    version_info["Server type"] = server_type.lower()

    try:
        pydantic_version = import_version("pydantic")
    except PackageNotFoundError:
        pydantic_version = "Not installed"

    version_info["Pydantic version"] = pydantic_version

    if connection_url_setting := get_current_settings().server.database.connection_url:
        import sqlalchemy as sa

        from prefect.server.database.dependencies import provide_database_interface
        from prefect.server.utilities.database import get_dialect

        connection_url = connection_url_setting.get_secret_value()
        database = get_dialect(connection_url).name
        version_info["Server"] = {"Database": database}
        if database == "sqlite":
            version_info["Server"]["SQLite version"] = sqlite3.sqlite_version
        elif database == "postgresql":
            try:
                db = provide_database_interface()
                async with db.session_context() as session:
                    result = await session.execute(sa.text("SHOW server_version"))
                    if postgres_version := result.scalar():
                        version_info["Server"]["PostgreSQL version"] = postgres_version
            except Exception:
                pass

    if not omit_integrations:
        integrations = _get_prefect_integrations()
        if integrations:
            version_info["Integrations"] = integrations

    _display(version_info)


def _get_prefect_integrations() -> dict[str, str]:
    """Get information about installed Prefect integrations."""
    from importlib.metadata import distributions

    integrations: dict[str, str] = {}
    for dist in distributions():
        name = dist.metadata.get("Name")
        if name and name.startswith("prefect-"):
            author_email = dist.metadata.get("Author-email", "").strip()
            if author_email.endswith("@prefect.io>"):
                if name == "prefect-client" and dist.version == "0.0.0":
                    continue
                integrations[name] = dist.version

    return integrations


def _display(object: dict[str, Any], nesting: int = 0) -> None:
    """Recursive display of a dictionary with nesting."""
    for key, value in object.items():
        key += ":"
        if isinstance(value, dict):
            _cli.console.print(" " * nesting + key)
            _display(value, nesting + 2)
        else:
            prefix = " " * nesting
            _cli.console.print(f"{prefix}{key.ljust(21 - len(prefix))} {value}")



"""
Dev command — native cyclopts implementation.

Internal Prefect development commands.
"""

import os
import platform
import shutil
import subprocess
import sys
import textwrap
import time
from functools import partial
from typing import Annotated, Optional

import cyclopts

import prefect.cli._app as _cli
from prefect.cli._utilities import (
    exit_with_error,
    exit_with_success,
    with_cli_exception_handling,
)

# Module-level import for test patchability
from prefect.utilities.processutils import run_process  # noqa: E402

DEV_HELP = """
Internal Prefect development.

Note that many of these commands require extra dependencies (such as npm and MkDocs)
to function properly.
"""

dev_app: cyclopts.App = cyclopts.App(
    name="dev",
    help=DEV_HELP,
    version_flags=[],
    help_flags=["--help"],
)


def _exit_with_error_if_not_editable_install() -> None:
    import prefect as _prefect

    if (
        _prefect.__module_path__.parent == "site-packages"
        or not (_prefect.__development_base_path__ / "pyproject.toml").exists()
    ):
        exit_with_error(
            "Development commands require an editable Prefect installation. "
            "Development commands require content outside of the 'prefect' module  "
            "which is not available when installed into your site-packages. "
            f"Detected module path: {_prefect.__module_path__}."
        )


@dev_app.command(name="build-docs")
@with_cli_exception_handling
def build_docs(schema_path: Optional[str] = None):
    """Builds REST API reference documentation for static display."""
    import json

    import prefect as _prefect

    _exit_with_error_if_not_editable_install()

    from prefect.server.api.server import create_app

    schema = create_app(ephemeral=True).openapi()

    if not schema_path:
        path_to_schema = (
            _prefect.__development_base_path__ / "docs" / "api-ref" / "schema.json"
        ).absolute()
    else:
        path_to_schema = os.path.abspath(schema_path)
    schema["info"] = {}
    with open(path_to_schema, "w") as f:
        json.dump(schema, f)
    _cli.console.print(f"OpenAPI schema written to {path_to_schema}")


@dev_app.command(name="build-ui")
@with_cli_exception_handling
def build_ui(no_install: bool = False):
    """Installs dependencies and builds UI locally. Requires npm."""
    import prefect as _prefect
    from prefect.utilities.filesystem import tmpchdir

    _exit_with_error_if_not_editable_install()
    with tmpchdir(_prefect.__development_base_path__ / "ui"):
        if not no_install:
            _cli.console.print("Installing npm packages...")
            try:
                subprocess.check_output(["npm", "ci"], shell=sys.platform == "win32")
            except Exception:
                _cli.console.print(
                    "npm call failed - try running `nvm use` first.", style="red"
                )
                raise
            _cli.console.print("Building for distribution...")
            env = os.environ.copy()
            subprocess.check_output(
                ["npm", "run", "build"], env=env, shell=sys.platform == "win32"
            )

    with tmpchdir(_prefect.__development_base_path__):
        if os.path.exists(_prefect.__ui_static_path__):
            _cli.console.print("Removing existing build files...")
            shutil.rmtree(_prefect.__ui_static_path__)

        _cli.console.print("Copying build into src...")
        shutil.copytree("ui/dist", _prefect.__ui_static_path__)

    _cli.console.print("Complete!")


@dev_app.command(name="ui")
@with_cli_exception_handling
async def ui():
    """Starts a hot-reloading development UI."""
    import prefect as _prefect
    from prefect.utilities.filesystem import tmpchdir

    _exit_with_error_if_not_editable_install()
    with tmpchdir(_prefect.__development_base_path__ / "ui"):
        _cli.console.print("Installing npm packages...")
        await run_process(["npm", "install"], stream_output=True)

        _cli.console.print("Starting UI development server...")
        await run_process(command=["npm", "run", "serve"], stream_output=True)


@dev_app.command(name="api")
@with_cli_exception_handling
async def api(
    *,
    host: Annotated[
        Optional[str],
        cyclopts.Parameter("--host", help="API host address."),
    ] = None,
    port: Annotated[
        Optional[int],
        cyclopts.Parameter("--port", help="API port number."),
    ] = None,
    log_level: Annotated[
        str,
        cyclopts.Parameter("--log-level", help="Log level for the server."),
    ] = "DEBUG",
    services: Annotated[
        bool,
        cyclopts.Parameter("--services/--no-services", help="Run services in app."),
    ] = True,
):
    """Starts a hot-reloading development API."""
    import signal

    import anyio
    import watchfiles

    import prefect as _prefect
    from prefect.settings import (
        PREFECT_SERVER_API_HOST,
        PREFECT_SERVER_API_PORT,
    )

    resolved_host = host if host is not None else PREFECT_SERVER_API_HOST.value()
    resolved_port = port if port is not None else PREFECT_SERVER_API_PORT.value()

    server_env = os.environ.copy()
    server_env["PREFECT_API_SERVICES_RUN_IN_APP"] = str(services)
    server_env["PREFECT_API_SERVICES_UI"] = "False"
    server_env["PREFECT_UI_API_URL"] = f"http://{resolved_host}:{resolved_port}/api"

    command = [
        sys.executable,
        "-m",
        "uvicorn",
        "--factory",
        "prefect.server.api.server:create_app",
        "--host",
        str(resolved_host),
        "--port",
        str(resolved_port),
        "--log-level",
        log_level.lower(),
    ]

    _cli.console.print(f"Running: {' '.join(command)}")

    stop_event = anyio.Event()
    start_command = partial(
        run_process, command=command, env=server_env, stream_output=True
    )

    async with anyio.create_task_group() as tg:
        try:
            server_pid = await tg.start(start_command)
            async for _ in watchfiles.awatch(
                _prefect.__module_path__,
                stop_event=stop_event,  # type: ignore
            ):
                _cli.console.print("Restarting Prefect Server...")
                os.kill(server_pid, signal.SIGTERM)  # type: ignore
                server_pid = await tg.start(start_command)
        except RuntimeError as err:
            if str(err).strip() != "Already borrowed":
                raise
        except KeyboardInterrupt:
            try:
                os.kill(server_pid, signal.SIGTERM)  # type: ignore
            except ProcessLookupError:
                pass
            stop_event.set()


@dev_app.command(name="start")
@with_cli_exception_handling
async def start(
    *,
    exclude_api: Annotated[
        bool,
        cyclopts.Parameter("--no-api", negative="", help="Exclude the API service."),
    ] = False,
    exclude_ui: Annotated[
        bool,
        cyclopts.Parameter("--no-ui", negative="", help="Exclude the UI service."),
    ] = False,
):
    """Starts a hot-reloading development server with API, UI, and agent processes."""
    import anyio

    from prefect.settings import (
        PREFECT_SERVER_API_HOST,
        PREFECT_SERVER_API_PORT,
    )

    async with anyio.create_task_group() as tg:
        if not exclude_api:
            tg.start_soon(
                partial(
                    api,
                    host=PREFECT_SERVER_API_HOST.value(),
                    port=PREFECT_SERVER_API_PORT.value(),
                )
            )
        if not exclude_ui:
            tg.start_soon(ui)


@dev_app.command(name="build-image")
@with_cli_exception_handling
def build_image(
    *,
    arch: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--arch",
            help=(
                "The architecture to build the container for. "
                "Defaults to the architecture of the host Python. "
                f"[default: {platform.machine()}]"
            ),
        ),
    ] = None,
    python_version: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--python-version",
            help=(
                "The Python version to build the container for. "
                "Defaults to the version of the host Python."
            ),
        ),
    ] = None,
    flavor: Annotated[
        Optional[str],
        cyclopts.Parameter(
            "--flavor",
            help=(
                "An alternative flavor to build, for example 'conda'. "
                "Defaults to the standard Python base image"
            ),
        ),
    ] = None,
    build_arg: Annotated[
        list[str],
        cyclopts.Parameter(
            "--build-arg",
            help=(
                "This will directly pass a --build-arg into the docker build process. "
                "Can be added to the command line multiple times."
            ),
        ),
    ] = [],
    dry_run: Annotated[
        bool,
        cyclopts.Parameter("--dry-run", help="Print the command instead of running."),
    ] = False,
):
    """Build a docker image for development."""
    import prefect as _prefect
    from prefect.utilities.dockerutils import (
        get_prefect_image_name,
        python_version_minor,
    )

    _exit_with_error_if_not_editable_install()

    arch = arch or platform.machine()
    python_version = python_version or python_version_minor()

    tag = get_prefect_image_name(python_version=python_version, flavor=flavor)

    command = [
        "docker",
        "build",
        str(_prefect.__development_base_path__),
        "--tag",
        tag,
        "--platform",
        f"linux/{arch}",
        "--build-arg",
        "PREFECT_EXTRAS=[dev]",
        "--build-arg",
        f"PYTHON_VERSION={python_version}",
    ]

    if flavor:
        command += ["--build-arg", f"BASE_IMAGE=prefect-{flavor}"]

    for arg in build_arg:
        command += ["--build-arg", arg]

    if dry_run:
        print(" ".join(command))
        return

    try:
        subprocess.check_call(command, shell=sys.platform == "win32")
    except subprocess.CalledProcessError:
        exit_with_error("Failed to build image!")
    else:
        exit_with_success(f"Built image {tag!r} for linux/{arch}")


@dev_app.command(name="container")
@with_cli_exception_handling
def container(
    *,
    bg: Annotated[
        bool,
        cyclopts.Parameter("--bg", help="Run in background."),
    ] = False,
    name: Annotated[
        str,
        cyclopts.Parameter("--name", help="Container name."),
    ] = "prefect-dev",
    api: Annotated[
        bool,
        cyclopts.Parameter("--api/--no-api", help="Start API in container."),
    ] = True,
    tag: Annotated[
        Optional[str],
        cyclopts.Parameter("--tag", help="Docker image tag."),
    ] = None,
):
    """Run a docker container with local code mounted and installed."""
    import docker
    from docker.models.containers import Container

    import prefect as _prefect
    from prefect.utilities.dockerutils import get_prefect_image_name

    _exit_with_error_if_not_editable_install()

    client = docker.from_env()

    containers = client.containers.list()
    container_names = {c.name for c in containers}
    if name in container_names:
        exit_with_error(
            f"Container {name!r} already exists. Specify a different name or stop "
            "the existing container."
        )

    blocking_cmd = "prefect dev api" if api else "sleep infinity"
    tag = tag or get_prefect_image_name()

    ctr: Container = client.containers.create(
        image=tag,
        command=[
            "/bin/bash",
            "-c",
            (
                "pip install -e /opt/prefect/repo\\[dev\\] && touch /READY &&"
                f" {blocking_cmd}"
            ),
        ],
        name=name,
        auto_remove=True,
        working_dir="/opt/prefect/repo",
        volumes=[f"{_prefect.__development_base_path__}:/opt/prefect/repo"],
        shm_size="4G",
    )

    print(f"Starting container for image {tag!r}...")
    ctr.start()

    print("Waiting for installation to complete", end="", flush=True)
    try:
        ready = False
        while not ready:
            print(".", end="", flush=True)
            result = ctr.exec_run("test -f /READY")
            ready = result.exit_code == 0
            if not ready:
                time.sleep(3)
    except BaseException:
        print("\nInterrupted. Stopping container...")
        ctr.stop()
        raise

    print(
        textwrap.dedent(
            f"""
            Container {ctr.name!r} is ready! To connect to the container, run:

                docker exec -it {ctr.name} /bin/bash
            """
        )
    )

    if bg:
        print(
            textwrap.dedent(
                f"""
                The container will run forever. Stop the container with:

                    docker stop {ctr.name}
                """
            )
        )
        return

    try:
        print("Send a keyboard interrupt to exit...")
        ctr.wait()
    except KeyboardInterrupt:
        pass
    finally:
        print("\nStopping container...")
        ctr.stop()



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, NoReturn

from docket import Perpetual

from prefect.logging import get_logger
from prefect.server.events import triggers
from prefect.server.services.base import RunInEphemeralServers, Service
from prefect.server.services.perpetual_services import perpetual_service
from prefect.server.utilities.messaging import Consumer, create_consumer
from prefect.server.utilities.messaging._consumer_names import (
    generate_unique_consumer_name,
)
from prefect.settings.context import get_current_settings
from prefect.settings.models.server.services import ServicesBaseSetting

if TYPE_CHECKING:
    import logging


logger: "logging.Logger" = get_logger(__name__)


class ReactiveTriggers(RunInEphemeralServers, Service):
    """Evaluates reactive automation triggers"""

    consumer_task: asyncio.Task[None] | None = None

    @classmethod
    def service_settings(cls) -> ServicesBaseSetting:
        return get_current_settings().server.services.triggers

    async def start(self) -> NoReturn:
        assert self.consumer_task is None, "Reactive triggers already started"
        consumer_name = generate_unique_consumer_name("reactive-triggers")
        logger.info(
            f"ReactiveTriggers starting with unique consumer name: {consumer_name}"
        )
        self.consumer: Consumer = create_consumer(
            "events",
            group="reactive-triggers",
            name=consumer_name,
            read_batch_size=self.service_settings().read_batch_size,
        )

        async with triggers.consumer() as handler:
            self.consumer_task = asyncio.create_task(self.consumer.run(handler))
            logger.debug("Reactive triggers started")

            try:
                await self.consumer_task
            except asyncio.CancelledError:
                pass

    async def stop(self) -> None:
        assert self.consumer_task is not None, "Reactive triggers not started"
        self.consumer_task.cancel()
        try:
            await self.consumer_task
        except asyncio.CancelledError:
            pass
        finally:
            await self.consumer.cleanup()
            self.consumer_task = None
        logger.debug("Reactive triggers stopped")


@perpetual_service(
    enabled_getter=lambda: get_current_settings().server.services.triggers.enabled,
    run_in_ephemeral=True,
)
async def evaluate_proactive_triggers_periodic(
    perpetual: Perpetual = Perpetual(
        automatic=True,
        every=get_current_settings().server.events.proactive_granularity,
    ),
) -> None:
    """Evaluate proactive automation triggers on a periodic schedule."""
    await triggers.evaluate_proactive_triggers()



"""
Shell command — native cyclopts implementation.

Run shell commands as Prefect flows.
"""

import logging
import subprocess
import sys
import threading
from typing import IO, Annotated, Any, Callable, Optional

import cyclopts

import prefect.cli._app as _cli
from prefect.cli._utilities import (
    exit_with_error,
    with_cli_exception_handling,
)

shell_app: cyclopts.App = cyclopts.App(
    name="shell", help="Serve and watch shell commands as Prefect flows."
)


def output_stream(pipe: IO[str], logger_function: Callable[[str], None]) -> None:
    """Read from a pipe line by line and log using the provided logging function."""
    with pipe:
        for line in iter(pipe.readline, ""):
            logger_function(line.strip())


def output_collect(pipe: IO[str], container: list[str]) -> None:
    """Collect output from a subprocess pipe and store it in a container list."""
    for line in iter(pipe.readline, ""):
        container.append(line)


@__import__("prefect").flow
def run_shell_process(
    command: str,
    log_output: bool = True,
    stream_stdout: bool = False,
    log_stderr: bool = False,
    popen_kwargs: Optional[dict[str, Any]] = None,
):
    """Execute a shell command and log its output.

    Designed for use within Prefect flows to run shell commands as part of
    task execution.

    Args:
        command: The shell command to execute.
        log_output: If True, log stdout/stderr to Prefect logs.
        stream_stdout: If True, stream stdout to Prefect logs.
        log_stderr: If True, log stderr to Prefect logs.
        popen_kwargs: Additional keyword arguments for subprocess.Popen.
    """
    from prefect.logging.loggers import get_run_logger

    logger = get_run_logger() if log_output else logging.getLogger("prefect")

    kwargs = {
        "stdout": subprocess.PIPE,
        "stderr": subprocess.PIPE,
        "shell": True,
        "text": True,
        "bufsize": 1,
        "universal_newlines": True,
    }

    if popen_kwargs:
        kwargs |= popen_kwargs

    stdout_container, stderr_container = [], []
    with subprocess.Popen(command, **kwargs) as proc:
        if stream_stdout:
            stdout_logger = logger.info
            output = output_stream
        else:
            stdout_logger = stdout_container
            output = output_collect

        stdout_thread = threading.Thread(
            target=output, args=(proc.stdout, stdout_logger)
        )
        stderr_thread = threading.Thread(
            target=output_collect, args=(proc.stderr, stderr_container)
        )

        stdout_thread.start()
        stderr_thread.start()

        stdout_thread.join()
        stderr_thread.join()

        proc.wait()
        if stdout_container:
            logger.info("".join(stdout_container).strip())

        if stderr_container and log_stderr:
            logger.error("".join(stderr_container).strip())
        if proc.returncode != 0:
            logger.error("".join(stderr_container).strip())
            sys.tracebacklimit = 0
            from prefect.exceptions import FailedRun

            raise FailedRun(f"Command failed with exit code {proc.returncode}")


@shell_app.command(name="watch")
@with_cli_exception_handling
async def watch(
    command: str,
    *,
    log_output: Annotated[
        bool,
        cyclopts.Parameter(
            "--log-output", negative="--no-log-output", help="Log output to Prefect."
        ),
    ] = True,
    flow_run_name: Annotated[
        Optional[str],
        cyclopts.Parameter("--flow-run-name", help="Name of the flow run."),
    ] = None,
    flow_name: Annotated[
        str,
        cyclopts.Parameter("--flow-name", help="Name of the flow."),
    ] = "Shell Command",
    stream_stdout: Annotated[
        bool,
        cyclopts.Parameter(
            "--stream-stdout", negative="--no-stream-stdout", help="Stream output."
        ),
    ] = True,
    tag: Annotated[
        Optional[list[str]],
        cyclopts.Parameter("--tag", help="Tags for the flow run (repeatable)."),
    ] = None,
):
    """
    Execute a shell command and observe it as a Prefect flow.
    """
    from prefect.context import tags

    tag = (tag or []) + ["shell"]

    defined_flow = run_shell_process.with_options(
        name=flow_name, flow_run_name=flow_run_name
    )
    with tags(*tag):
        defined_flow(
            command=command, log_output=log_output, stream_stdout=stream_stdout
        )


@shell_app.command(name="serve")
@with_cli_exception_handling
async def serve(
    command: str,
    *,
    flow_name: Annotated[
        str,
        cyclopts.Parameter("--flow-name", help="Name of the flow."),
    ],
    deployment_name: Annotated[
        str,
        cyclopts.Parameter("--deployment-name", help="Name of the deployment."),
    ] = "CLI Runner Deployment",
    deployment_tags: Annotated[
        Optional[list[str]],
        cyclopts.Parameter("--tag", help="Deployment tags (repeatable)."),
    ] = None,
    log_output: Annotated[
        bool,
        cyclopts.Parameter("--log-output", help="Log command output.", show=False),
    ] = True,
    stream_stdout: Annotated[
        bool,
        cyclopts.Parameter(
            "--stream-stdout", negative="--no-stream-stdout", help="Stream output."
        ),
    ] = True,
    cron_schedule: Annotated[
        Optional[str],
        cyclopts.Parameter("--cron-schedule", help="Cron schedule."),
    ] = None,
    timezone: Annotated[
        Optional[str],
        cyclopts.Parameter("--timezone", help="Timezone for the schedule."),
    ] = None,
    concurrency_limit: Annotated[
        Optional[int],
        cyclopts.Parameter("--concurrency-limit", help="Max concurrent flow runs."),
    ] = None,
    run_once: Annotated[
        bool,
        cyclopts.Parameter("--run-once", help="Run once instead of forever."),
    ] = False,
):
    """
    Create and serve a deployment that runs a shell command.
    """
    if concurrency_limit is not None and concurrency_limit < 1:
        exit_with_error("--concurrency-limit must be >= 1.")

    from prefect.client.schemas.actions import DeploymentScheduleCreate
    from prefect.client.schemas.schedules import CronSchedule
    from prefect.runner import Runner
    from prefect.settings import get_current_settings
    from prefect.types.entrypoint import EntrypointType

    schedule = (
        CronSchedule(cron=cron_schedule, timezone=timezone) if cron_schedule else None
    )
    defined_flow = run_shell_process.with_options(name=flow_name)

    runner_deployment = await defined_flow.to_deployment(
        name=deployment_name,
        parameters={
            "command": command,
            "log_output": log_output,
            "stream_stdout": stream_stdout,
        },
        entrypoint_type=EntrypointType.MODULE_PATH,
        schedules=[DeploymentScheduleCreate(schedule=schedule)] if schedule else [],
        tags=(deployment_tags or []) + ["shell"],
    )

    runner = Runner(name=flow_name)
    deployment_id = await runner.add_deployment(runner_deployment)
    help_message = (
        f"[green]Your flow {runner_deployment.flow_name!r} is being served and polling"
        " for scheduled runs!\n[/]\nTo trigger a run for this flow, use the following"
        " command:\n[blue]\n\t$ prefect deployment run"
        f" '{runner_deployment.flow_name}/{deployment_name}'\n[/]"
    )
    if ui_url := get_current_settings().ui_url:
        help_message += (
            "\nYou can also run your flow via the Prefect UI:"
            f" [blue]{ui_url}/deployments/deployment/{deployment_id}[/]\n"
        )

    _cli.console.print(help_message, soft_wrap=True)
    await runner.start(run_once=run_once)
